#!/usr/bin/env python3 

import os
import numpy as np
import pandas as pd

from numpy import count_nonzero

from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import seaborn as sns
import matplotlib.pyplot as plt

from utils import median_of_ratios_normalisation, calculate_percentile, extract_samples_to_condition


###################
## Class definition 
###################

class MetaboliteAnalysis:
    '''
    A class to streamline the filtering and exploration of a metabolome dataset.   


    Parameters
    ----------
    metabolome_csv: str
        A path to a .csv file with the metabolome data (scaled or unscaled).
        Shape of the dataframe is usually (n_samples, n_features) with n_features >> n_samples
        
    metabolome_feature_id_col: str, optional
        The name of the column that contains the feature identifiers (default is 'feature_id').
        Feature identifiers should be unique (=not duplicated).

    
    Attributes
    ----------
    metabolome: `pandas.core.frame.DataFrame`, (n_samples, n_features)
      The metabolome Pandas dataframe imported from the .csv file. 
    metabolome_validated: `bool`
      Is the metabolome dataset validated?
      Default is False.
    blank_features_filtered: `bool`
      Are the features present in blank samples filtered out from the metabolome data?
      Default by False.
    normalised_by_median_of_ratios_method: bool
      Are the feature peak areas normalised by the median of ratios method (from DESeq2)?
      Default is False.
    filtered_by_percentile_value: bool
      Are the features filtered by percentile value?
    unreliable_features_filtered: `bool`
      Are the features not reliably present within one group filtered out from the metabolome data?
    pca_performed: `bool`
      Has PCA been performed on the metabolome data?
      Default is False. 
    exp_variance: `pandas.core.frame.DataFrame`, (n_pc, 1)
      A Pandas dataframe with explained variance per Principal Component.
      The index of the df contains the PC index (PC1, PC2, etc.).
      The second column contains the percentage of the explained variance per PC.
    metabolome_pca_reduced: `numpy.ndarray`, (n_samples, n_pc)
      Numpy array with sample coordinates in reduced dimensions.
      The dimension of the numpy array is the minimum of the number of samples and features. 
    sparsity: float
      Metabolome matrix sparsity.


    Methods
    -------
    validate_input_metabolome_df
      Check if the provided metabolome file is suitable. Turns attribute metabolome_validated to True. 
    discard_features_detected_in_blanks
      Removes features only detected in blank samples. 
    filter_out_unreliable_features()
      Filter out features not reliably detectable in replicates of the same grouping factor. 
      For instance, if a feature is detected less than 4 times within 4 biological replicates.  
    write_clean_metabolome_to_csv()
      Write the filtered and analysis-ready metabolome data to a .csv file. 
       
   
    Notes
    -----
    Example of an input metabolome input format (from a csv file)

    +----------------------+---------+---------+---------+---------+-------+-------+-------+-------+----------+----------+----------+----------+
    | feature_id           | blank_1 | blank_2 | blank_3 | blank_4 | MM_1  | MM_2  | MM_3  | MM_4  | LA1330_1 | LA1330_2 | LA1330_3 | LA1330_4 |
    +======================+=========+=========+=========+=========+=======+=======+=======+=======+==========+==========+==========+==========+
    | rt-0.04_mz-241.88396 | 280     | 694     | 502     | 604     | 554   | 678   | 674   | 936   | 824      | 940      | 794      | 828      |
    +----------------------+---------+---------+---------+---------+-------+-------+-------+-------+----------+----------+----------+----------+
    | rt-0.05_mz-143.95911 | 1036    | 1566    | 1326    | 1490    | 1364  | 1340  | 1692  | 1948  | 1928     | 1956     | 1730     | 1568     |
    +----------------------+---------+---------+---------+---------+-------+-------+-------+-------+----------+----------+----------+----------+
    | rt-0.06_mz-124.96631 | 1308    | 992     | 1060    | 1010    | 742   | 990   | 0     | 888   | 786      | 668      | 762      | 974      |
    +----------------------+---------+---------+---------+---------+-------+-------+-------+-------+----------+----------+----------+----------+
    | rt-0.08_mz-553.45905 | 11340   | 12260   | 10962   | 11864   | 10972 | 11190 | 12172 | 11820 | 12026    | 11604    | 11122    | 11260    |
    +----------------------+---------+---------+---------+---------+-------+-------+-------+-------+----------+----------+----------+----------+
    | rt-0.08_mz-413.26631 | 984     | 1162    | 1292    | 1104    | 1090  | 1106  | 1290  | 1170  | 1282     | 924      | 1172     | 1062     |
    +----------------------+---------+---------+---------+---------+-------+-------+-------+-------+----------+----------+----------+----------+

    
    Example
    -------
    >>> met = MetaboliteAnalysis(
        metabolome_csv='my_metabolome_data.csv', 
        metabolome_feature_id_col='feature_id')
    >>> met.validate_input_metabolome_df()
    Metabolome input data validated


    See also
    --------
    scikit-learn PCA: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
    
    '''
    ### Class attributes
    # By default the metabolome and phenotype data imported from .csv files will have to be validated
    # By default all filters have not been executed (blank filtering, etc.)
    metabolome_validated = False
    blank_features_filtered = False
    normalised_by_median_of_ratios_method = False
    filtered_by_percentile_value = False
    unreliable_features_filtered = False
    pca_performed = False
    sparsity=None


    ##########################
    # Class constructor method
    ##########################
    def __init__(
        self, 
        metabolome_csv, 
        metabolome_feature_id_col='feature_id'):
        """
        Constructor method. 
        Returns a Python instance of class MetabolomeAnalysis 
        """     
        # Import metabolome dataframe and verify presence of feature id column
        self.metabolome = pd.read_csv(metabolome_csv)
        if metabolome_feature_id_col not in self.metabolome.columns:
            raise ValueError("The specified column with feature identifiers {0} is not present in your '{1}' file.".format(metabolome_feature_id_col,os.path.basename(metabolome_csv)))
        else:
            self.metabolome.set_index(metabolome_feature_id_col, inplace=True)
    
    def validate_input_metabolome_df(self, metabolome_feature_id_col='feature_id'):
        '''
        Validates the dataframe containing the feature identifiers, metabolite values and sample names.
        Will place the 'feature_id_col' column as the index of the validated dataframe. 
        The validated metabolome dataframe is stored as the 'validated_metabolome' attribute. 
        
        Parameters
        ----------
        metabolome_feature_id: str, optional 
            The name of the column that contains the feature identifiers (default is 'feature_id').
            Feature identifiers should be unique (=not duplicated).
            
        Returns
        -------
        self: object
          Object with attribute metabolome_validated set to True if tests are passed. 
        
        Notes
        -----
        Example of a valid input metabolome dataframe


                      | genotypeA_rep1 | genotypeA_rep2 | genotypeA_rep3 | genotypeA_rep4 |
                      |----------------|----------------|----------------|----------------|
          feature_id
        | metabolite1 |   1246         | 1245           | 12345          | 12458          |
        | metabolite2 |   0            | 0              | 0              | 0              |
        | metabolite3 |   10           | 0              | 0              | 154            |
        
        '''

        if np.any(self.metabolome.values < 0):
            raise ValueError("Sorry, metabolite values have to be zero or positive integers (>=0)")
        else:
            print("Metabolome input data validated.")
            self.metabolome_validated = True

    #############################################
    ### Filter features detected in blank samples
    #############################################
    def discard_features_detected_in_blanks(
        self, 
        blank_sample_contains='blank'):
        '''
        Removes 
        Steps:
          1. Sum the abundance of each feature in the blank samples.
          2. Makes a list of features to be discarded (features with a positive summed abundance).
          3. Returns a filtered Pandas dataframe with only features not detected in blank samples

        Parameters
        ----------
        blank_sample_contains: str, optional.
            Column names with this name will be considered blank samples.
            Default is='blank'
        
        Returns
        -------
        metabolome: pandas.core.frame.DataFrame
            A filtered Pandas dataframe without features detected in blank samples and with the blank samples removed. 
        '''
        if self.metabolome_validated == True:
            pass
        else:
            self.validate_input_metabolome_df()
        blank_cols = [col for col in self.metabolome.columns.values.tolist() if blank_sample_contains in col]
        # If the sum of a feature in blank samples is higher than 0 then 
        # this feature should be removed
        # only keep features that are not detectable in blank samples
        self.metabolome["sum_features"] = self.metabolome[blank_cols].sum(axis=1)
        self.metabolome = self.metabolome.query("sum_features == 0")
        # Remove columns with blank samples and the sum column used for filtering
        self.metabolome = self.metabolome.drop(blank_cols, axis=1)
        self.metabolome = self.metabolome.drop("sum_features", axis=1)    

    
    ########################################
    ### Normalise by median of ratios method
    ########################################
    def normalise_values_by_median_of_ratios_method(self):
        '''
        Use DESeq2 median of ratios method to normalise feature peak area values.

        Returns
        -------
        self: object
          Object with metabolome values normalised and median_of_ratios_normalised attribute set to True. 
        
        Example
        -------
        >>> met = MetaboliteAnalysis(
            metabolome_csv='tests/metabolome_test_data.csv', 
            metabolome_feature_id_col='feature_id')
        >>> met.validate_input_metabolome_df()
        Metabolome input data validated
        >>> met.normalise_values_by_median_of_ratios_method()
        >>> met.normalised_by_median_of_ratios_method
        True

        See also
        --------
        Median of ratios method explained:
          - https://scienceparkstudygroup.github.io/rna-seq-lesson/05-descriptive-plots/index.html#54-generate-normalized-counts
          - https://hbctraining.github.io/DGE_workshop/lessons/02_DGE_count_normalization.html 
        '''
        if self.normalised_by_median_of_ratios_method == True:
            print("Data is already normalised (median of ratios)")
        else:
            df_to_normalize = self.metabolome
            normalised_df = median_of_ratios_normalisation(df_to_normalize)
            self.metabolome = normalised_df
            self.normalised_by_median_of_ratios_method = True

    #######################################################################
    # Create density plots of feature peak areas for each grouping variable
    #######################################################################
    def create_density_plot(self, name_grouping_var="genotype", n_cols=3, nbins=1000):
        '''
        For each grouping variable (e.g. genotype), creates a histogram and density plot of all feature peak areas.
        This plot helps to see whether some groups have a value distribution different from the rest. 
        The percentage is indicated on the y-axis (bar heights sum to 100).
        
        Parameters
        ----------
        name_grouping_var: str, optional
            The name used when splitting between replicate and main factor.
            For example "genotype" when splitting MM_rep1 into 'MM' and 'rep1'.
            Default is 'genotype'. 
        n_cols: int, optional
            The number of columns for the final plot.
        nbins: int, optional
            The number of bins to create. 
        
        Returns
        -------
        matplotlib Axes
            Returns the Axes object with the density plots drawn onto it.
        '''
        df = self.metabolome
        melted_df = pd.melt(
            df.reset_index(), 
            id_vars=df.index.name, 
            var_name='sample')
        samples2conditions = extract_samples_to_condition(df)
        melted_df_with_cond = melted_df.merge(samples2conditions, on='sample')
        fig = plt.figure()
        g = sns.FacetGrid(melted_df_with_cond, col='genotype', col_wrap=3)
        g = g.map_dataframe(sns.histplot, x='value', kde=True, stat='percent', bins=nbins)
        g.set_titles(col_template="{col_name}")
        g.set_xlabels("Peak area value (AU, log scale)")
        g.set_ylabels("Percentage of total (%)")
        plt.xscale('log')

    
    ###########################################################
    ### Filter features that are lower than a certain threshold
    ###########################################################
    def filter_features_per_group_by_percentile(
        self, 
        name_grouping_var="genotype",
        separator_replicates="_",
        percentile=50):
        '''
        Filter metabolome dataframe based on a selected percentile threshold.
        Features with a peak area values lower than the selected percentile will be discarded. 
        The percentile value is calculated per grouping variable. 

        For instance, selecting the 50th percentile (median) will discard 50% of the features with a peak area
        lower than the median/50th percentile in each group. 

        Parameters
        ----------
        name_grouping_var: str, optional
            The name of the grouping variable (default is "genotype")
        separator_replicates: str, optional
            The character used to separate the main grouping variable from biological replicates. 
            Default is "_: (underscore)
        percentile: float, optional
            The percentile threshold. Has to be comprised 0 and 100.

        Returns
        -------
        self: object
            The object with the .metabolome attribute filtered and the filtered_by_percentile_value set to True. 

        Example
        -------
        >>> met = MetaboliteAnalysis(
            metabolome_csv='tests/metabolome_test_data.csv', 
            metabolome_feature_id_col='feature_id')
        >>> met.validate_input_metabolome_df()
        Metabolome input data validated
        >>> met.discard_features_detected_in_blanks(blank_sample_contains="blank")
        >> met.metabolome.shape
        (7544, 32)
        >>> met.filter_features_based_on_peak_area_level(percentile=90)
        >>> met.metabolome.shape
        (3171, 32)

        
        See also
        --------
        create_density_plot() method to decide on a suitable percentile value. 
        '''
        df = self.metabolome
        my_percentile_threshold = calculate_percentile(
            df, 
            my_percentile=percentile)
        # melt to tidy/long format (one value per row, one column per variable)
        # add grouping variable name
        melted_df = pd.melt(
            df.reset_index(), 
            id_vars='feature_id', 
            var_name='sample', 
            value_name='value')
        sample2condition = extract_samples_to_condition(df, name_grouping_var=name_grouping_var, separator_replicates=separator_replicates)
        melted_df_with_group = melted_df.merge(sample2condition, on="sample")
        
        # calculate selected percentile value per group 
        percentiles_per_group = melted_df_with_group.groupby('genotype').apply(lambda df: calculate_percentile(df["value"], my_percentile=percentile))

        # extract features per group which abundance is strictly higher than the selected percentile value
        # do that for each group e.g. each genotype
        # deduplicate the final list of features to keep 
        features_to_keep = []
        all_groups = percentiles_per_group.index.values
        feature_id_colname = df.index.name
        for my_group in all_groups:
            sub_df = melted_df_with_group[melted_df_with_group[name_grouping_var] == my_group]
            my_group_percentile = percentiles_per_group.loc[my_group]
            sub_df_filtered = sub_df.loc[sub_df["value"] > my_group_percentile]
            good_features = sub_df_filtered[feature_id_colname]
            for good_feature in good_features:
                features_to_keep.append(good_feature)
        features_to_keep = list(set(features_to_keep)) # deduplicate
        df_filtered = df.loc[features_to_keep,:]

        self.metabolome = df_filtered
        self.filtered_by_percentile_value = True


    #######################################################################################
    ### Filter features that are not reliable (detected less than nb replicates in a group)
    ######################################################################################
    def filter_out_unreliable_features(
        self,
        name_grouping_var="genotype", 
        nb_times_detected=4,
        separator_replicates='_'):
        '''
        Removes features not reliably detectable in multiple biological replicates from the same grouping factor. 

        Takes a dataframe with feature identifiers in index and samples as columns.
        Step 1: First melt and split the sample names to generate the grouping variable
        Step 2: count number of times a metabolite is detected in the groups. 
        If number of times detected in a group = number of biological replicates then it is considered as reliable
        Each feature receives a tag  'reliable' or 'not_reliable'
        Step 3: discard the 'not_reliable' features and keep the filtered dataframe. 

        Params
        ------
        name_grouping_var: str, optional
            The name used when splitting between replicate and main factor.
            For example "genotype" when splitting MM_rep1 into 'MM' and 'rep1'.
            Default is 'genotype'. 
        nb_times_detected: int, optionaldefault=4
            Number of times a metabolite should be detected to be considered 'reliable'. 
            Should be equal to the number of biological replicates for a given group of interest (e.g. genotype)
        separator_replicates: string, default="_"
            The separator to split sample names into a grouping variable (e.g. genotype) and the biological replicate number (e.g. 1)
        

        Returns
        -------
        metabolome: ndarray
            A Pandas dataframe with only features considered as reliable, sample names and their values. 
        
        Notes 
        -----
        Input dataframe

                             	| MM_1  	| MM_2  	| MM_3  	| MM_4  	| LA1330_1 	| LA1330_2 	|
                            	|----------	|----------	|----------	|----------	|----------	|----------	|
          feature_id           	 
        | rt-0.04_mz-241.88396 	| 554   	| 678   	| 674   	| 936   	| 824      	| 940      	|
        | rt-0.05_mz-143.95911 	| 1364  	| 1340  	| 1692  	| 1948  	| 1928     	| 1956     	|
        | rt-0.06_mz-124.96631 	| 0      	| 0     	| 0     	| 888   	| 786      	| 668      	|
        | rt-0.08_mz-553.45905 	| 10972 	| 11190 	| 12172 	| 11820 	| 12026    	| 11604    	|

        Output df (rt-0.06_mz-124.96631 is kicked out because 3x0 and 1x888 in MM groups)
        
                             	| MM_1  	| MM_2  	| MM_3  	| MM_4  	| LA1330_1 	| LA1330_2 	|
                            	|----------	|----------	|----------	|----------	|----------	|----------	|
          feature_id           	 
        | rt-0.04_mz-241.88396 	| 554   	| 678   	| 674   	| 936   	| 824      	| 940      	|
        | rt-0.05_mz-143.95911 	| 1364  	| 1340  	| 1692  	| 1948  	| 1928     	| 1956     	|
        | rt-0.08_mz-553.45905 	| 10972 	| 11190 	| 12172 	| 11820 	| 12026    	| 11604    	|


        '''
        df = self.metabolome

        ### Melt (required to tag reliable/not reliable features)
        melted_df = pd.melt(
            df.reset_index(), 
            id_vars='feature_id', 
            var_name="sample")
        melted_df[[name_grouping_var, 'rep']] = melted_df["sample"].str.split(pat=separator_replicates, expand=True)
        sorted_melted_df = melted_df.sort_values(by=['feature_id', name_grouping_var])
        sorted_melted_df.set_index('feature_id', inplace=True)
        sorted_melted_df_parsed = sorted_melted_df.drop('sample', axis=1)
        
        ### Identify features that are reliable
        # Creates a dictionary to have the feature identifier and a 'reliable'/'not_reliable' tag
        reliability_feature_dict = {}
        features = sorted_melted_df_parsed.index.unique().tolist()

        for feature in features:
          # Dataframe containing only one metabolite
          temp_df = sorted_melted_df_parsed.loc[feature,:]
          temp_df = temp_df.drop('rep', axis=1)

          # number of values above 0 across all groups
          min_number_of_values_above_0_across_all_groups = temp_df[temp_df['value'] > 0].groupby(by=name_grouping_var)[name_grouping_var].count().max()

          # If the feature is detected a minimum of times equal to the number of biological replicates
          # This means the feature is reliably detectable in at least one group (e.g. one genotype)
          if min_number_of_values_above_0_across_all_groups >= nb_times_detected:
            reliability_feature_dict[feature] = "reliable"    
          else:
            reliability_feature_dict[feature] = 'not_reliable'

        # Convert to a Pandas to prepare the filtering of the feature/abundance dataframe
        reliability_df = pd.DataFrame.from_dict(
            reliability_feature_dict, orient="index", columns=["reliability"]).reset_index()
        
        reliability_df = reliability_df.rename({"index":"feature_id"}, axis='columns')
        reliability_df = reliability_df[reliability_df["reliability"] == "reliable"]
        features_to_keep = reliability_df.feature_id.tolist()

        df_reliable_features = df.loc[features_to_keep,:]
                
        self.metabolome = df_reliable_features
        self.unreliable_features_filtered = True

    #################################################
    ### Write filtered metabolomoe data to a csv file
    #################################################

    def write_clean_metabolome_to_csv(self, path_of_cleaned_csv="filtered_metabolome.csv"):
        '''
        A function that verify that the metabolome dataset has been cleaned up. 
        Writes the metabolome data as a comma-separated value file on disk

        Parameters
        ----------
        path_of_cleaned_csv: str, optional
            The path and filename of the .csv file to save.
            Default to "./filtered_metabolome.csv" 
        '''
        try:
            self.blank_features_filtered == True
        except:
            raise ValueError("Features in blank should be removed first using the 'discard_features_detected_in_blanks() method.")
        
        try:
            self.unreliable_features_filtered == True
        except:
            raise ValueError("Features not reliably detected within at least one group should be removed first using the 'filter_out_unreliable_features() method.") 
        
        self.metabolome.to_csv(path_or_buf=path_of_cleaned_csv, sep=',')



    #################################
    ## Principal Component Analysis
    #################################

    def compute_pca_on_metabolites(self, scale=True, n_principal_components=10, auto_transpose=True):
        """
        Performs a Principal Component Analysis (PCA) on the metabolome data. 
        
        The PCA analysis will return transformed coordinates of the samples in a new space. 
        It will also give the percentage of variance explained by each Principal Component. 
        Assumes that number of samples < number of features/metabolites
        Performs a transpose of the metabolite dataframe if n_samples > n_features (this can be turned off with auto_transpose)
        
        Parameters
        ----------
        scale: `bool`, optional
            Perform scaling (standardize) the metabolite values to zero mean and unit variance. 
            Default is True. 
        n_principal_components: int, optional
            number of principal components to keep in the PCA analysis.
            if number of PCs > min(n_samples, n_features) then set to the minimum of (n_samples, n_features)
            Default is to calculate 10 components.
        auto_transpose: `bool`, optional. 
            If n_samples > n_features, performs a transpose of the feature matrix.
            Default is True (meaning that transposing will occur if n_samples > n_features).
    
        Returns
        -------
        self: object
          Object with .exp_variance: dataframe with explained variance per Principal Component
          .metabolome_pca_reduced: dataframe with samples in reduced dimensions
          .pca_performed: `bool`ean set to True
        """
        # Verify that samples are in rows and features in columns
        # Usually n_samples << n_features so we should have n_rows << n_cols
        n_rows = self.metabolome.shape[0]
        n_cols = self.metabolome.shape[1]
        
        metabolite_df = self.metabolome
        if n_rows > n_cols:
            # Likely features are in row so transpose to have samples in rows
            metabolite_df = metabolite_df.transpose()
        else:
            pass
    
        if scale == True:
            scaler = StandardScaler(with_mean=True, with_std=True)
            metabolite_df_scaled = scaler.fit_transform(metabolite_df)
            metabolite_df_scaled = pd.DataFrame(metabolite_df_scaled)
            metabolite_df_scaled.columns = metabolite_df.columns
            metabolite_df_scaled.set_index(metabolite_df.index.values)       
        else:
            pass

        if n_principal_components <= np.minimum(n_rows, n_cols):
            pca = PCA(n_components=n_principal_components)
            metabolite_df_scaled_transformed = pca.fit_transform(metabolite_df_scaled)
            exp_variance = pd.DataFrame(pca.explained_variance_ratio_.round(2)*100, columns=["explained_variance"])
        # If n_principal_components > min(n_samples, n_features)
        # then n_principal_components = min(n_samples, n_features)
        else:
            n_principal_components > np.minimum(n_rows, n_cols)
            n_principal_components = np.minimum(n_rows, n_cols)
            pca = PCA(n_components=n_principal_components)
            metabolite_df_scaled_transformed = pca.fit_transform(metabolite_df_scaled)
            exp_variance = pd.DataFrame(pca.explained_variance_ratio_.round(2)*100, columns=["explained_variance"])         

        # The numbering of the components starts by default at 0. 
        # Setting this to 1 to make it more user friendly
        exp_variance.index = exp_variance.index+1
        
        # Store PCA results 
        self.exp_variance = exp_variance
        self.metabolome_pca_reduced = metabolite_df_scaled_transformed
        self.pca_performed = True

    def create_scree_plot(self, plot_file_name=None):
        '''
        Returns a barplot with the explained variance per Principal Component. 
        Has to be preceded by perform_pca()

        Parameters
        ---------
        plot_file_name: string, default='None'
          Path to a file where the plot will be saved.
          For instance 'my_scree_plot.pdf'

        Returns
        -------
        matplotlib Axes
            Returns the Axes object with the scree plot drawn onto it.
            Optionally a saved image of the plot. 
        '''
        try:
            self.pca_performed
        except: 
            raise AttributeError("Please compute the PCA first using the compute_pca_on_metabolites() method.") 
        
        sns.barplot(
            data=self.exp_variance, 
            x=self.exp_variance.index, 
            y="explained_variance")
        plt.xlabel("Principal Component")
        plt.ylabel("Explained variance (%)")

        # Optionally save the plot
        if plot_file_name != None:
            plot_dirname = os.path.dirname(plot_file_name)
            if plot_dirname == '': # means file will be saved in current working directory
                plt.savefig(plot_file_name)
            else:
                os.makedirs(plot_dirname, exist_ok=True)
                plt.savefig(plot_file_name)

        plt.show()


    def create_sample_score_plot(
        self, 
        pc_x_axis=1, 
        pc_y_axis=2, 
        name_grouping_var='genotype',
        separator_replicates="_",
        plot_file_name=None):
        '''
        Returns a sample score plot of the samples on PCx vs PCy. 
        Samples are colored based on the grouping variable (e.g. genotype)

        Parameters
        ----------
        pc_x_axis: int, optional 
          Principal Component to plot on the x-axis (default is 1 so PC1 will be plotted).
        pc_y_axis: int, optional.
           Principal Component to plot on the y-axis (default is 2 so PC2 will be plotted).
        name_grouping_var: str, optional
          Name of the variable used to color samples (Default is "genotype"). 
        separator_replicates: str, optional.
          String separator that separates grouping factor from biological replicates (default is underscore "_").
        plot_file_name: str, optional 
          A file name and its path to save the sample score plot (default is None).
          For instance "mydir/sample_score_plot.pdf"
          Path is relative to current working directory.
        
        Returns
        -------
        matplotlib Axes
            Returns the Axes object with the sample score plot drawn onto it.
            Samples are colored by specified grouping variable. 
            Optionally a saved image of the plot. 


        '''
        if self.pca_performed:
            pass
        else:
            raise AttributeError("Please compute the PCA first using the compute_pca_on_metabolites() method.") 

        n_features = self.metabolome.shape[0]
        n_samples = self.metabolome.shape[1]
        min_of_samples_and_features = np.minimum(n_samples, n_features)
        
        samples_to_conditions = extract_samples_to_condition(df=self.metabolome, name_grouping_var=name_grouping_var, separator_replicates=separator_replicates)

        if pc_x_axis == pc_y_axis:
            raise ValueError("Values for Principal Components on x axis and y axis have to be different.")
        if pc_x_axis > min_of_samples_and_features:
            raise ValueError("Your principal component for x axis should be lower than {0}".format(min_of_samples_and_features))
        if pc_y_axis > np.minimum(n_samples, n_features):
            raise ValueError("Your principal component for y axis should be lower than {0}".format(min_of_samples_and_features))
       
        
        if not name_grouping_var in samples_to_conditions.columns:
            raise IndexError("The grouping variable '{0}' is not present in the samples_to_condition dataframe".format(name_grouping_var))
        else:
            # Build the plot
            plt.figure(figsize=(10,7))
            
            self.scatter_plot = sns.scatterplot(
            x=self.metabolome_pca_reduced[:,pc_x_axis-1],
            y=self.metabolome_pca_reduced[:,pc_y_axis-1],
            hue=samples_to_conditions[name_grouping_var],
            s=200)

            plt.xlabel("PC" + str(pc_x_axis) + ": " + str(self.exp_variance.iloc[pc_x_axis-1,0].round(2)) + "% variance") 
            plt.ylabel("PC" + str(pc_y_axis) + ": " + str(self.exp_variance.iloc[pc_y_axis-1,0].round(2)) + "% variance")
            plt.title("PC" + str(pc_x_axis) + " vs PC" + str(pc_y_axis))

            # Optionally save the plot
            if plot_file_name != None:
                plot_dirname = os.path.dirname(plot_file_name)
                if plot_dirname == '': # means file will be saved in current working directory
                    plt.savefig(plot_file_name)
                else:
                    os.makedirs(plot_dirname, exist_ok=True)
                    plt.savefig(plot_file_name)

            plt.show()


    #######################################################################################
    ### Determine sparsity (number of non-zero)
    ######################################################################################
    def compute_metabolome_sparsity(self):
        '''
        Determine the sparsity of the metabolome matrix. 
        Formula: number of non zero values/number of values * 100
        The higher the sparsity, the more zero values 
        
        Returns
        -------
        self: object
            Object with sparsity attribute filled (sparsity is a float).

        References
        ----------
        https://stackoverflow.com/questions/38708621/how-to-calculate-percentage-of-sparsity-for-a-numpy-array-matrix
        '''
        number_of_non_zero_values = count_nonzero(self.metabolome)
        total_number_of_values = self.metabolome.size
        sparsity = number_of_non_zero_values/total_number_of_values
        print("Sparsity of the metabolome matrix is equal to {0:.3f}".format(sparsity))
        self.sparsity=sparsity


#!/usr/bin/env python3 

import os
from warnings import WarningMessage
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import balanced_accuracy_score, precision_score, recall_score, f1_score, confusion_matrix, ConfusionMatrixDisplay

def median_of_ratios_normalisation(_data : pd.DataFrame) -> pd.DataFrame:
    """
    Normalize a dataframe with the median of ratios method 
    from DESeq2. 
    
   
    input data (as a pandas dataframe), e.g.:
    
            sample1    sample2    sample3
    gene1   0.00000    10.0000    4.00000
    gene2   2.00000    6.00000    12.0000
    gene3   33.5000    55.0000    200.000
    
    normalized output:
            sample1    sample2    sample3
    gene1   0.00000    10.6444    1.57882
    gene2   4.76032    6.38664    4.73646
    gene3   78.5453    58.5442    78.9410
    
    References
    ----------
    StatQuest: https://www.youtube.com/watch?v=UFB993xufUU
    HBC Harvard: https://hbctraining.github.io/DGE_workshop/lessons/02_DGE_count_normalization.html

    """
    # step 1: log normalize
    log_data = np.log(_data)

    # step 2: average rows
    row_avg = np.mean(log_data, axis=1)
    
    # step 3: filter rows with zeros
    rows_no_zeros = row_avg[row_avg != -np.inf].index
    
    # step 4: subtract avg log counts from log counts
    ratios = log_data.loc[rows_no_zeros].subtract(row_avg.loc[rows_no_zeros], axis=0)
    
    # step 5: calculate median of ratios
    medians = ratios.median(axis=0)
    
    # step 6: median -> base number
    scaling_factors = np.e ** medians
    
    # step 7: normalize!
    normalized_data = _data / scaling_factors
    return normalized_data

def calculate_percentile(df, my_percentile=50):
    '''
    Compute the q-th percentile of data.
    Returns the q-th percentile of the array elements.

    Parameters
    ----------
    my_percentile: float, optional
        Percentile which must be between 0 and 100.
      
    See also
    ---------
    numpy.percentile()
    https://numpy.org/doc/stable/reference/generated/numpy.percentile.html
    '''
    my_array = df.to_numpy()
    percentile_of_df = np.percentile(my_array, my_percentile)
    return percentile_of_df

def compute_metrics_classification(y_predictions, y_trues, positive_class):
    '''
    Compute a series of metrics for classification tasks

    Util function designed to work downstream of the search for the best model. 
    Will compute the following metrics:
      - balanced accuracy
      - precision
      - recall
      - f1 score

    Parameters
    ----------
    y_predictions: list
      List of class predictions. 
    y_trues: list
      List of the true values (from the test set)
    positive_class: str
      The name of the positive class for calculation of true positives, true negatives, etc. 

    Returns
    -------
    model_metrics_df: `pandas.core.frame.DataFrame`
      Dataframe with the balanced accuracy, precision, recall and f1 score calculated. 

    See also
    --------
    https://scikit-learn.org/stable/modules/model_evaluation.html
    '''
    
    balanced_accuracy = balanced_accuracy_score(y_pred=y_predictions, y_true=y_trues)
    precision = precision_score(y_pred=y_predictions, y_true=y_trues, pos_label=positive_class)
    recall = recall_score(y_pred=y_predictions, y_true=y_trues, pos_label=positive_class)
    f1 = f1_score(y_true=y_trues, y_pred=y_predictions, pos_label=positive_class)
    
    model_metrics_dict = {"balanced_accuracy": balanced_accuracy, "precision": precision, "recall": recall, "f1 score": f1}
    model_metrics_df = pd.DataFrame.from_dict(model_metrics_dict, orient="index", columns=["value"])
    model_metrics_df_rounded = model_metrics_df.round(3)

    return model_metrics_df_rounded 

def plot_confusion_matrix(y_predictions, y_trues):
    '''
    Plot confusion matrix

    Parameters
    ----------
    y_predictions: list
      List of class predictions. 
    y_trues: list
      List of the true values (from the test set)
    positive_class: str
      The name of the positive class for calculation of true positives, true negatives, etc. 

    Returns
    -------
    model_metrics_df: `pandas.core.frame.DataFrame`
      Dataframe with the balanced accuracy, precision, recall and f1 score calculated. 

    See also
    --------
    https://scikit-learn.org/stable/modules/model_evaluation.html
    '''
    cm = confusion_matrix(y_true=y_trues, y_pred=y_predictions)
    disp = ConfusionMatrixDisplay(
        confusion_matrix=cm)
    disp.plot()
    plt.show()


def extract_samples_to_condition(df, name_grouping_var='genotype', separator_replicates='_'):
    '''
    A utility function to extract the grouping factor (e.g. 'genotype') from sample names. 
    
    Uses melting (wide to long) and split grouping variable from biological replicates using specified separator.
    
    Parameters
    ----------
    df: pandas.core.DataFrame
    name_grouping_var: str, optional 
        Name of the variable used as grouping variable (default is 'genotype').
    separator_replicates: str, optional
        The separator between the grouping variable and the biological replicates ( default is underscore '_')
    
    Returns
    -------
    A dataframe with the correspondence between samples and experimental condition (grouping variable).

    Notes
    -------
    Input dataframe
                        | genotypeA_rep1 | genotypeA_rep2 | genotypeA_rep3 | genotypeA_rep4 |
                        |----------------|----------------|----------------|----------------|
          feature_id
        | metabolite1   |   1246         | 1245           | 12345          | 12458          |
        | metabolite2   |   0            | 0              | 0              | 0              |
        | metabolite3   |   10           | 0              | 0              | 154            |
    
    Output dataframe
        
        | sample             | genotype       | replicate      |
        |--------------------|----------------|----------------|
        | genotypeA_rep1     |   genotypeA    | rep1           |
        | genotypeA_rep2     |   genotypeA    | rep2           |
        | genotypeA_rep3     |   genotypeA    | rep3           |
        | genotypeA_rep4     |   genotypeA    | rep4           |
        | etc.
    '''
    melted_df = pd.melt(df.reset_index(), id_vars='feature_id', var_name="sample")
    melted_df[[name_grouping_var, 'rep']] = melted_df["sample"].str.split(pat=separator_replicates, expand=True)
    melted_df_parsed = melted_df.drop(["feature_id", "value"], axis=1)
    melted_df_parsed_dedup = melted_df_parsed.drop_duplicates()
    return melted_df_parsed_dedup



#!/usr/bin/env python3 

import os
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import scipy.optimize as opt
from numpy import arange


class PhenotypeAnalysis:
    '''
    A class to analyse data from developmental bioassays and group the samples in distict phenotypic classes. 
    - What does this class do? 

    Parameters
    ----------
    The constructor method __init__ takes one arguments:
    bioassay_csv: string
        A path to a .csv file with the bioassay count data.
        Shape of the dataframe is usually ...

    Attributes
    ----------
    metabolome_validated: boolean, default=False
      Is the metabolome dataset validated? 
    phenotype_validated=False
    blank_features_filtered=False
    unreliable_features_filtered=False

    '''
    def __init__(
        self, 
        bioassay_csv):

        # Import bioassay dataframe 
        self.bioassay = pd.read_csv(bioassay_csv)



    def reshape_to_wide(
        self,
        sample_id='sample_id',
        grouping_variable='genotype',
        developmental_stages='stage',
        count_values='number',
        time='day'):
        
        '''
        Reshapes the dataframe from a long to a wide format to make the data accessible for pre-processing.
        with the counts of each developmental stage in a seperate columns.
        
        Parameters
        ----------
        sample_id: string, default='sample_id'
            The name of the column that contains the sample identifiers.
        grouping_variable: string, default='genotype'
            The name of the column that contains the names of the grouping variables.
            Examples are genotypes or treatments
        developmental_stages: string, default='stage'
            The name of the column that contains the developmental stages that were scored during the bioassay.
        count_values: string, default='numbers'
            The name of the column that contains the counts.
        time: string, default='day'
            The name of the column that contains the time at which bioassay scoring was performed.
            Examples are the date or the number of days after infection.
        
        Examples
        --------
        Example of an input dataframe

        | sample_id | genotype  | day   | stage         | count |
        |-----------|-----------|-------|---------------|-------|
        | mm_1      |   mm      | 5     | eggs          | 45    |
        | mm_1      |   mm      | 5     | first_instar  | 0     |
        | mm_1      |   mm      | 5     | second_instar | 0     |
        
        
        Example of a reshaped output dataframe

        | sample_id | genotype  | day   | eggs  | first_instar  | second instar | third_instar  |
        |-----------|-----------|-------|-------|---------------|---------------|---------------|
        | mm_1      |   mm      | 5     | 45    | 0             | 0             | 0             |
        | mm_1      |   mm      | 9     | NA    | 10            | 5             | 0             |
        | mm_1      |   mm      | 11    | NA    | 15            | 17            | 4             |
        '''

        
        # check if specified columns exist in dataframe
        if sample_id not in self.bioassay.columns:
            raise ValueError("The specified column with sample identifiers {0} is not present in your '{1}' file.".format(sample_id,os.path.basename(bioassay_csv)))
        else:
            self.sample_id = sample_id

        if grouping_variable not in self.bioassay.columns:
            raise ValueError("The specified column with grouping variable names {0} is not present in your '{1}' file.".format(grouping_variable,os.path.basename(bioassay_csv)))
        else:
            pass

        if developmental_stages not in self.bioassay.columns:
            raise ValueError("The specified column with developmental stages {0} is not present in your '{1}' file.".format(developmental_stages,os.path.basename(bioassay_csv)))
        else:
            pass

        if count_values not in self.bioassay.columns:
            raise ValueError("The specified column with values {0} is not present in your '{1}' file.".format(count_values,os.path.basename(bioassay_csv)))
        else:
            pass

        if time not in self.bioassay.columns:
            raise ValueError("The specified column with time values (e.g. days after infection) {0} is not present in your '{1}' file.".format(time,os.path.basename(bioassay_csv)))
        else:
            pass

        # reshape the dataframe to a wide format with one developmental stage per column
        self.bioassay = self.bioassay.pivot(index=[sample_id, grouping_variable, time], columns=developmental_stages, values=count_values)
        self.bioassay = self.bioassay.reset_index()
        


    def combine_seperately_counted_versions_of_last_recorded_stage(
        self,
        exuviea='exuviea',
        late_last_stage='late_fourth_instar',
        early_last_stage='early_fourth_instar',
        new_last_stage='fourth_instar',
        seperate_exuviea=True,
        late_last_stage_removed=True,
        early_last_stage_kept=True,
        remove_individual_stage_columns=True):
        
        '''
        Calculates the total number of nymphs developed to the final developmental stage per sample on each timepoint.
        This is used when nymphs in the (late) final nymph stage were removed after each counting moment and/or
        when exuviea and last instar stage nymphs were counted seperately.
        Removal of late last stage nymphs could for example be used to prevent adults from emerging and escaping.
        
        Parameters
        ----------
        exuviea: string, default='exuviea'
            The name of the column that contains the exuviea counts. 
        late_last_stage: string, default='late_fourth_instar'
            The name of the column that contains the counts of the last developmental stage recorded in the bioassay.
        early_last_stage: string, default='early_fourth_instar'
            The name of the column that contains the counts of the nymphs in early last developmental stage.
            Is used when nymphs counted in late_last_stage were removed after each counting moment during the bioassay.
        new_last_stage: string, default='fourth_instar'
            Name for new column with the returned total final stage data
        seperate_exuviea: boolean, default=True
            If True, sums exuviea and late_last_stage per sample per timepoint.
            If exuviea were counted seperately from late_last_stage, set to True.
            If exuviea count was included in late_last_stage, set to False
        late_last_stage_removed: boolean, default=True
            If True, returns the cumulative number of late_last_stage(+exuviea) per sample over time.
            If nymphs counted in late_last_stage (and exuviea if counted seperately) were removed after each counting 
            moment, set to True.
            If nymphs counted in late_last_stage (and exuviea if counted seperately) were left on the sample until
            ending the bioassay, set to False.
       early_last_stage_kept: boolean, default=True
            If True, sums the early and late last stage counts per sample per timepoint
            If late last stage nymphs were removed after each counting moment, but early last stage nymphs were left on
            sample, set to True.
            If early and late last stage nymphs were not counted seperately, set to False
        remove_individual_stage_columns: boolean, default=True
            If True, removes exuviea, late_last_stage, early_last_stage columns from dataframe after returning 
            new_last_stage column.
        
        Examples
        --------
        Example of an input dataframe

        | sample_id | genotype  | day   | eggs  | ... | third_instar  | exuviea   | early_fourth_instar | late_fourth_instar |
        |-----------|-----------|-------|-------|-----|---------------|-----------|---------------------|--------------------|
        | mm_1      |   mm      | 5     | 45    | ... | 0             | 0         | 0                   | 0                  |
        | mm_1      |   mm      | 9     | NA    | ... | 0             | 1         | 5                   | 0                  |
        | mm_1      |   mm      | 11    | NA    | ... | 4             | 0         | 7                   | 4                  |
        
        
        Example of an output dataframe

        | sample_id | genotype  | day   | eggs  | first_instar  | second instar | third_instar  | fourth_instar |
        |-----------|-----------|-------|-------|---------------|---------------|---------------|---------------|
        | mm_1      |   mm      | 5     | 45    | 0             | 0             | 0             | 0             |
        | mm_1      |   mm      | 9     | NA    | 10            | 5             | 0             | 6
        | mm_1      |   mm      | 11    | NA    | 15            | 17            | 4             | 12
        '''

        # if the exuviea and late last stage nymphs were counted individualy but treated simmilarly,
        # the exuviea and late last stage nymphs should be summed before going further
        if seperate_exuviea == True:
            
            # check if specified columns with exuviea and late last stage are present
            if exuviea not in self.bioassay.columns:
                raise ValueError("The specified column with exuviea counts {0} is not present in your file.".format(exuviea))
            else:
                pass

            if late_last_stage not in self.bioassay.columns:
                raise ValueError("The specified column with late last stage counts {0} is not present in your file.".format(late_last_stage))
            else:
                pass

            # if specified columns are present, sum exuviea and late last stage
            self.bioassay['late_exuviea'] = self.bioassay[[exuviea, late_last_stage]].sum(axis=1)
        else:
            if late_last_stage not in self.bioassay.columns:
                raise ValueError("The specified column with late last stage counts {0} is not present in your file.".format(late_last_stage))
            else:
                pass
            self.bioassay['late_exuviea'] = self.bioassay[late_last_stage]

        # if the late last stage nymphs were removed after each count, the cumulative number should be used
        # when analysing the development to this stage
        if late_last_stage_removed == True:
            self.bioassay['late_exuviea'] = self.bioassay.groupby([self.sample_id])['late_exuviea'].cumsum()
        else:
            pass

        # if the early last stage nymphs were kept for further development after counting, the early and late last stage
        # nymphs should be combined for the total number of last stage nymphs
        if early_last_stage_kept == True:

            # check if specified column with early last stage counts is present 
            if early_last_stage not in self.bioassay.columns:
                raise ValueError("The specified column with early last stage counts {0} is not present in your file.".format(early_last_stage))
            else:
                pass

            # check if specified column name for total last instar numbers is not yet present 
            if new_last_stage in self.bioassay.columns:
                raise ValueError("The specified column name for total last instar numbers {0} already exists in your file.".format(new_last_stage))
            else:
                pass

            # if only the early last stage column is present, calculate total last stage nymphs
            self.bioassay[new_last_stage] = self.bioassay[['late_exuviea', early_last_stage]].sum(axis=1)
        else:

            # check if specified column name for total last instar numbers is not yet present 
            if new_last_stage in self.bioassay.columns:
                raise ValueError("The specified column name for total last instar numbers {0} already exists in your file.".format(new_last_stage))
            else:
                pass
            self.bioassay[new_last_stage] = self.bioassay['late_exuviea']

        # cleaning up unwanted columns
        if remove_individual_stage_columns == True:
            self.bioassay = self.bioassay.drop(columns=['late_exuviea'])
            self.bioassay = self.bioassay.drop([exuviea, late_last_stage, early_last_stage], axis=1)
        else:
            pass


    def correct_cumulative_counts(
        self, 
        current_stage,
        grouping_variable):
        '''
        Inner function for convert_counts_to_cumulative(). If nymphs die during the bioassay, 
        they should be included in the cumulative count for the stages it had passed. 
        Otherwise, the cumulative count could go down over time. This function corrects the cumulative
        count if it is lower than the previous count.
        '''

        grouped_df = self.bioassay.groupby(grouping_variable)
        corrected_df = pd.DataFrame()

        for day, group in grouped_df:
            temp_df = group
            temp_df = temp_df.reset_index()
            temp_df['test'] = temp_df[current_stage]

            for i in range(1, len(temp_df)):
                if temp_df.loc[i-1, 'test'] != 0:
                    if temp_df.loc[i, 'test'] < temp_df.loc[i-1, 'test']:
                        temp_df.loc[i, 'test'] = temp_df.loc[i-1, 'test']
                    else:
                        pass
                else:
                    pass

            corrected_df = pd.concat([corrected_df, temp_df], ignore_index=True)

        self.bioassay = corrected_df
        self.bioassay[current_stage] = self.bioassay['test']
        self.bioassay = self.bioassay.drop(columns=['test', 'index'])


    def create_df_with_max_counts_per_stage(
        self, 
        egg_column,
        last_stage,
        grouping_variable):
        '''
        Inner function for convert_counts_to_cumulative(). 
        With the maximum number of nymphs developed to or past each developmental stage per plant, 
        making graphs becomes easier.
        '''

        grouped_df = self.bioassay.groupby(grouping_variable)
        self.max_counts = pd.DataFrame()

        for day, group in grouped_df:
            temp_df = group
            temp_df = temp_df.reset_index()
            temp2_df = pd.DataFrame()
            temp2_df = temp_df.nlargest(1, [last_stage])
            temp2_df[egg_column] = temp_df[egg_column].max()

            self.max_counts = pd.concat([self.max_counts, temp2_df], ignore_index=True)

        self.max_counts = self.max_counts.drop(columns=['index'])



    def convert_counts_to_cumulative(
        self,
        n_developmental_stages=4,
        sample_id='sample_id',
        eggs='eggs',
        first_stage='first_instar',
        second_stage='second_instar',
        third_stage='third_instar',
        fourth_stage='fourth_instar',
        fifth_stage='fifth_instar',
        sixth_stage='sixth_instar'):
        
        '''
        Calculates the total number of nymphs developed to or past each stage on each timepoint.
        Cumulative counts make the analysis of development over time and the comparison of number of nymphs past a stage easier.
        If nymphs in the (late) final nymph stage were removed after each counting moment and/or
        when exuviea and/or early and late last instar stage nymphs were counted seperately, 
        total_last_stage() should be used first.
        
        Parameters
        ----------
        n_developmental_stages: integer, default=4
            The number of developmental stage swhich were recorded seperately. 
            Can range from 2 till 6.
        sample_id: string, default='sample_id'
            The name of the column that contains the sample identifiers.
        eggs: string, default='eggs'
            The name of the column that contains the counts of the eggs.
        first_stage: string, default='first_instar'
            The name of the column that contains the counts of the first developmental stage recorded in the bioassay.
        second_stage: string, default='second_instar'
            The name of the column that contains the counts of the second developmental stage recorded in the bioassay.
        third_stage: string, default='third_instar'
            The name of the column that contains the counts of the third developmental stage recorded in the bioassay.
        fourth_stage: string, default='fourth_instar'
            The name of the column that contains the counts of the fourth developmental stage recorded in the bioassay.
        fifth_stage: string, default='fifth_instar'
            The name of the column that contains the counts of the fifth developmental stage recorded in the bioassay.
        sixth_stage: string, default='sixth_instar'
            The name of the column that contains the counts of the sixth developmental stage recorded in the bioassay.
        '''
        
        # check if specified column with sample_id is present 
        if sample_id not in self.bioassay.columns:
            raise ValueError("The specified column with unique sample identifiers {0} is not present in your file.".format(sample_id))
        else:
            pass

        if n_developmental_stages == 1:
            # check if specified columns with counts per stage are present 
            if first_stage not in self.bioassay.columns:
                raise ValueError("The specified column with first stage counts {0} is not present in your file.".format(first_stage))
            else:
                pass


            self.correct_cumulative_counts(current_stage=first_stage, grouping_variable=sample_id)

            self.create_df_with_max_counts_per_stage(egg_column=eggs, last_stage=first_stage, grouping_variable=sample_id)
            
        
        elif n_developmental_stages == 2:

            # check if specified columns with counts per stage are present 
            if first_stage not in self.bioassay.columns:
                raise ValueError("The specified column with first stage counts {0} is not present in your file.".format(first_stage))
            else:
                pass
            if second_stage not in self.bioassay.columns:
                raise ValueError("The specified column with second stage counts {0} is not present in your file.".format(second_stage))
            else:
                pass
            

            # if all specified columns are present, first calculate the cumulative numbers for all stages before 
            # correcting the cumulative counts. Doing it otherwise might increase the risk of double counts.
            self.bioassay[first_stage] = self.bioassay[[first_stage, second_stage]].sum(axis=1)

            self.correct_cumulative_counts(current_stage=second_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=first_stage, grouping_variable=sample_id)

            self.create_df_with_max_counts_per_stage(egg_column=eggs, last_stage=second_stage, grouping_variable=sample_id)

        
        elif n_developmental_stages == 3:
            
            # check if specified columns with counts per stage are present 
            if first_stage not in self.bioassay.columns:
                raise ValueError("The specified column with first stage counts {0} is not present in your file.".format(first_stage))
            else:
                pass
            if second_stage not in self.bioassay.columns:
                raise ValueError("The specified column with second stage counts {0} is not present in your file.".format(second_stage))
            else:
                pass
            if third_stage not in self.bioassay.columns:
                raise ValueError("The specified column with third stage counts {0} is not present in your file.".format(third_stage))
            else:
                pass

            # if all specified columns are present, first calculate the cumulative numbers for all stages before 
            # correcting the cumulative counts. Doing it otherwise might increase the risk of double counts.
            self.bioassay[first_stage] = self.bioassay[[first_stage, second_stage, third_stage]].sum(axis=1)
            self.bioassay[second_stage] = self.bioassay[[second_stage, third_stage]].sum(axis=1)

            self.correct_cumulative_counts(current_stage=first_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=second_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=third_stage, grouping_variable=sample_id)

            self.create_df_with_max_counts_per_stage(egg_column=eggs, last_stage=third_stage, grouping_variable=sample_id)


        
        
        elif n_developmental_stages == 4:

            # check if specified columns with counts per stage are present 
            if first_stage not in self.bioassay.columns:
                raise ValueError("The specified column with first stage counts {0} is not present in your file.".format(first_stage))
            else:
                pass
            if second_stage not in self.bioassay.columns:
                raise ValueError("The specified column with second stage counts {0} is not present in your file.".format(second_stage))
            else:
                pass
            if third_stage not in self.bioassay.columns:
                raise ValueError("The specified column with third stage counts {0} is not present in your file.".format(third_stage))
            else:
                pass
            if fourth_stage not in self.bioassay.columns:
                raise ValueError("The specified column with fourth stage counts {0} is not present in your file.".format(fourth_stage))
            else:
                pass

            # if all specified columns are present, first calculate the cumulative numbers for all stages before 
            # correcting the cumulative counts. Doing it otherwise might increase the risk of double counts.
            self.bioassay[first_stage] = self.bioassay[[first_stage, second_stage, third_stage, fourth_stage]].sum(axis=1)
            self.bioassay[second_stage] = self.bioassay[[second_stage, third_stage, fourth_stage]].sum(axis=1)
            self.bioassay[third_stage] = self.bioassay[[third_stage, fourth_stage]].sum(axis=1)

            self.correct_cumulative_counts(current_stage=first_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=second_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=third_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=fourth_stage, grouping_variable=sample_id)

            self.create_df_with_max_counts_per_stage(egg_column=eggs, last_stage=fourth_stage, grouping_variable=sample_id)


        elif n_developmental_stages == 5:

            # check if specified columns with counts per stage are present 
            if first_stage not in self.bioassay.columns:
                raise ValueError("The specified column with first stage counts {0} is not present in your file.".format(first_stage))
            else:
                pass
            if second_stage not in self.bioassay.columns:
                raise ValueError("The specified column with second stage counts {0} is not present in your file.".format(second_stage))
            else:
                pass
            if third_stage not in self.bioassay.columns:
                raise ValueError("The specified column with third stage counts {0} is not present in your file.".format(third_stage))
            else:
                pass
            if fourth_stage not in self.bioassay.columns:
                raise ValueError("The specified column with fourth stage counts {0} is not present in your file.".format(fourth_stage))
            else:
                pass
            if fifth_stage not in self.bioassay.columns:
                raise ValueError("The specified column with fifth stage counts {0} is not present in your file.".format(fifth_stage))
            else:
                pass

            # if all specified columns are present, first calculate the cumulative numbers for all stages before 
            # correcting the cumulative counts. Doing it otherwise might increase the risk of double counts.
            self.bioassay[first_stage] = self.bioassay[[first_stage, second_stage, third_stage, fourth_stage, fifth_stage]].sum(axis=1)
            self.bioassay[second_stage] = self.bioassay[[second_stage, third_stage, fourth_stage, fifth_stage]].sum(axis=1)
            self.bioassay[third_stage] = self.bioassay[[third_stage, fourth_stage, fifth_stage]].sum(axis=1)
            self.bioassay[fourth_stage] = self.bioassay[[fourth_stage, fifth_stage]].sum(axis=1)

            self.correct_cumulative_counts(current_stage=first_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=second_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=third_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=fourth_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=fifth_stage, grouping_variable=sample_id)

            self.create_df_with_max_counts_per_stage(egg_column=eggs, last_stage=fifth_stage, grouping_variable=sample_id)


        elif n_developmental_stages == 6:

            # check if specified columns with counts per stage are present 
            if first_stage not in self.bioassay.columns:
                raise ValueError("The specified column with first stage counts {0} is not present in your file.".format(first_stage))
            else:
                pass
            if second_stage not in self.bioassay.columns:
                raise ValueError("The specified column with second stage counts {0} is not present in your file.".format(second_stage))
            else:
                pass
            if third_stage not in self.bioassay.columns:
                raise ValueError("The specified column with third stage counts {0} is not present in your file.".format(third_stage))
            else:
                pass
            if fourth_stage not in self.bioassay.columns:
                raise ValueError("The specified column with fourth stage counts {0} is not present in your file.".format(fourth_stage))
            else:
                pass
            if fifth_stage not in self.bioassay.columns:
                raise ValueError("The specified column with fifth stage counts {0} is not present in your file.".format(fifth_stage))
            else:
                pass
            if sixth_stage not in self.bioassay.columns:
                raise ValueError("The specified column with sixth stage counts {0} is not present in your file.".format(sixth_stage))
            else:
                pass

            # if all specified columns are present, first calculate the cumulative numbers for all stages before 
            # correcting the cumulative counts. Doing it otherwise might increase the risk of double counts.
            self.bioassay[first_stage] = self.bioassay[[first_stage, second_stage, third_stage, fourth_stage, fifth_stage, sixth_stage]].sum(axis=1)
            self.bioassay[second_stage] = self.bioassay[[second_stage, third_stage, fourth_stage, fifth_stage, sixth_stage]].sum(axis=1)
            self.bioassay[third_stage] = self.bioassay[[third_stage, fourth_stage, fifth_stage, sixth_stage]].sum(axis=1)
            self.bioassay[fourth_stage] = self.bioassay[[fourth_stage, fifth_stage, sixth_stage]].sum(axis=1)
            self.bioassay[fifth_stage] = self.bioassay[[fifth_stage, sixth_stage]].sum(axis=1)

            self.correct_cumulative_counts(current_stage=first_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=second_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=third_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=fourth_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=fifth_stage, grouping_variable=sample_id)
            self.correct_cumulative_counts(current_stage=sixth_stage, grouping_variable=sample_id)

            self.create_df_with_max_counts_per_stage(egg_column=eggs, last_stage=sixth_stage, grouping_variable=sample_id)

    
    def plot_counts_per_stage(
        self, 
        grouping_variable='genotype',
        sample_id='sample_id',
        eggs='eggs',
        first_stage='first_instar',
        second_stage='second_instar',
        third_stage='third_instar',
        fourth_stage='fourth_instar',
        make_nymphs_relative_to='first_instar'):
        '''
        Plots the counts per nymphal stage in boxplots. The nymph counts are given as the absolute number of nymphs that 
        developed to or past each stage at the last timepoint and as a fraction of nymphs that developed to or past each 
        stage at the last timepoint relative to another developmental stage. The other developmental stage to which the 
        data is made relative defaults to the first instar stage, because this represents the number of hatched eggs. This
        means that in this case only the succes of the development is compared between groups (e.g. genotypes or 
        treatments) and the hatching rate of the eggs is not taken into acount.

        The imput dataframe 'max_counts' is created with convert_counts_to_cumulative.
        
        Parameters
        ----------
        grouping_variable: string, default='genotype'
            The name of the column that contains the names of the grouping variables.
            Examples are genotypes or treatments
        sample_id: string, default='sample_id'
            The name of the column that contains the sample identifiers.
        eggs: string, default='eggs'
            The name of the column that contains the counts of the eggs.
        first_stage: string, default='first_instar'
            The name of the column that contains the counts of the first developmental stage recorded in the bioassay.
        second_stage: string, default='second_instar'
            The name of the column that contains the counts of the second developmental stage recorded in the bioassay.
        third_stage: string, default='third_instar'
            The name of the column that contains the counts of the third developmental stage recorded in the bioassay.
        fourth_stage: string, default='fourth_instar'
            The name of the column that contains the counts of the fourth developmental stage recorded in the bioassay.
        make_nymphs_relative_to: string, default='first_instar'
            The name of the column that contains the counts of the developmental stage which should be used to calculate 
            therelative development to all developmental stages.
        
        Examples
        --------
        Example of an input dataframe

        | sample_id | genotype  | day   | eggs  | first_instar  | second_instar | third_instar | fourth_instar |
        |-----------|-----------|-------|-------|---------------|---------------|--------------|---------------|
        | mm_1      |   mm      | 28    | 45    | 34            | 30            | 30           | 29            |
        | mm_2      |   mm      | 28    | 50    | 39            | 33            | 28           | 26            |
        | LA_1      |   LA      | 28    | 42    | 30            | 25            | 17           | 4             |
        
        '''
        
        self.absolute_counts = pd.DataFrame()
        self.absolute_counts = pd.melt(self.max_counts, id_vars=[sample_id, grouping_variable], 
                value_vars=[eggs, first_stage, second_stage, third_stage, fourth_stage], 
                var_name='developmental_stage', value_name='absolute_count')
        
        plots = sns.FacetGrid(self.absolute_counts, col='developmental_stage')
        plots.map(sns.boxplot, grouping_variable, 'absolute_count', palette="colorblind")
        plots.set(ylim=(0, None))
        
        
        self.max_relative = pd.DataFrame()
        self.max_relative = self.max_counts
        self.max_relative['hatching_rate'] = self.max_relative[first_stage]/self.max_relative[eggs]
        self.max_relative[second_stage] = self.max_relative[second_stage]/self.max_relative[make_nymphs_relative_to]
        self.max_relative[third_stage] = self.max_relative[third_stage]/self.max_relative[make_nymphs_relative_to]
        self.max_relative[fourth_stage] = self.max_relative[fourth_stage]/self.max_relative[make_nymphs_relative_to]

        self.max_relative = pd.melt(self.max_relative, id_vars=[sample_id, grouping_variable], 
                value_vars=['hatching_rate', second_stage, third_stage, fourth_stage], 
                var_name='developmental_stage', value_name='relative_count')
        
        plots = sns.FacetGrid(self.max_relative, col='developmental_stage')
        plots.map(sns.boxplot, grouping_variable, 'relative_count', palette="colorblind")
        plots.set(ylim=(0,1))

        ## I want to add statistics here using an ANOVA with a Dunnett's post-hoc, but at this moment (Nov 1 2022) a Dunnett's
        ## test is not available on python yet. 



    def plot_development_over_time_in_fitted_model(
        self, 
        sample_id='sample_id',
        grouping_variable='genotype',
        time='day',
        stage_of_ineterest='fourth_instar',
        use_relative_data=True,
        make_nymphs_relative_to='first_instar'):
        '''
        Fits a 3 parameter log-logistic curve to the development over time to a specified stage. The fitted curve and the
        observed datapoints are plotted and returned with the model parameters. 
        The reduced Chi-squared is provided to asses the goodness of fit for the fitted models for each group (genotype, 
        treatment, etc.). Optimaly, the reduced Chi-squared should approach the number of observation points per sample. A
        much larger reduced Chi-squared indicates a bad fit. A much smaller reduced Chi-squared indicates overfitting of 
        the model.
        
        Parameters
        ----------
        sample_id: string, default='sample_id'
            The name of the column that contains the sample identifiers.
        grouping_variable: string, default='genotype'
            The name of the column that contains the names of the grouping variables.
            Examples are genotypes or treatments.
        time: string, default='day'
            The name of the column that contains the time at which bioassay scoring was performed.
            Examples are the date or the number of days after infection.
        stage_of_ineterest: string, default='fourth_instar'
            The name of the column that contains the data of the developmental stage of interest.
        use_relative_data: boolean, default=True
            If True, the counts for the stage of interest are devided by the stage indicated at 'make_nymphs_relative_to'.
            The returned relative rate is used for plotting and curve fitting.
        make_nymphs_relative_to: string, default='first_instar'
            The name of the column that contains the counts of the developmental stage which should be used to calculate 
            therelative development to all developmental stages.

        
        Examples
        --------
        Example of an input dataframe

        | sample_id | genotype  | day   | eggs  | first_instar  | second instar | third_instar  | fourth_instar |
        |-----------|-----------|-------|-------|---------------|---------------|---------------|---------------|
        | mm_1      |   mm      | 5     | 45    | 15            | 7             | 0             | 0             |
        | mm_1      |   mm      | 9     | NA    | 24            | 14            | 6             | 3             |
        | mm_1      |   mm      | 11    | NA    | 38            | 27            | 16            | 12            |
        
        '''

        # define function of model:
        def ll3(x,slope,maximum,emt50):
            ''' 
            A three parameter log-logistic function.
        
            Parameters
            ----------
            slope: 
                the slope of the curve
            maximum: 
                the maximum value of the curve
            emt50: 
                the EmT50, the timepoint at which 50% of nymphs has developed to the stage of interest
            '''
            return(maximum/(1+np.exp(slope*(np.log(x)-np.log(emt50)))))
        
        # extract the timecourse in which the bioassay was performed. Needed to fit the model
        x_line = arange(min(self.bioassay[time]), max(self.bioassay[time])+1, 1)

        # if relative counts should be used
        if use_relative_data==True:
            grouped_df = self.bioassay.groupby(sample_id)
            temp_df = pd.DataFrame()
            for name, group in grouped_df:
                temp2_df = group
                temp2_df = temp2_df.reset_index()
                temp2_df['relative_stage'] = temp2_df[stage_of_ineterest]/max(temp2_df[make_nymphs_relative_to])

                temp_df = pd.concat([temp_df, temp2_df], ignore_index=True)

            self.bioassay = temp_df
            self.bioassay = self.bioassay.drop(columns='index')

        else:
            self.bioassay['relative_stage'] = self.bioassay[stage_of_ineterest]


        # add a column with standard deviations to use for the sigma in the curve_fit function
        grouped_df = self.bioassay.groupby([grouping_variable,time])
        stdev_df = pd.DataFrame()
        for name, group in grouped_df:
            temp_df = group
            temp_df = temp_df.reset_index()
            temp_df['stdev'] = temp_df['relative_stage'].std()
            if temp_df['relative_stage'].std() == 0:
                temp_df['stdev'] = 10

            stdev_df = pd.concat([stdev_df, temp_df], ignore_index=True)

        self.bioassay = stdev_df
        self.bioassay = self.bioassay.drop(columns='index')

        # the model is fitted to the individual groups to obtain the parameters for each group:
        grouped_df = self.bioassay.groupby(grouping_variable)
        fit_df = []
        fitted_df = []
        for name,group in grouped_df:
            
            # make an initial guess of the parameters as if the data is linear
            p0 = [-(max(group['relative_stage'])/(max(self.bioassay[time])+1)), max(group['relative_stage']), (max(self.bioassay[time])+1)/2]
            
            # fit the model to the data
            popt, pcov = opt.curve_fit(ll3, group[time], group['relative_stage'], p0=p0)
            
            # store the model parameters with their standard deviations in a df
            temp_df = dict(zip(['slope', 'maximum', 'emt50'], popt))
            temp2_df = dict(zip(['slope_sd', 'maximum_sd', 'emt50_sd'], np.sqrt(np.diag(pcov))))
            temp_df['slope(±sd)'] = '%.2f' % temp_df['slope'] + "(±" + '%.2f' % temp2_df['slope_sd'] + ")"
            temp_df['maximum(±sd)'] = '%.2f' % temp_df['maximum'] + "(±" + '%.2f' % temp2_df['maximum_sd'] + ")"
            temp_df['emt50(±sd)'] = '%.2f' % temp_df['emt50'] + "(±" + '%.2f' % temp2_df['emt50_sd'] + ")"
            temp_df[grouping_variable] = name

            # calculate chi2 for goodness of fit of model
            residuals = group['relative_stage']-ll3(group[time],*popt)
            sq_residuals = residuals**2
            chi_sq = np.sum(sq_residuals / group['stdev']**2)
            temp_df['reduced_chi2'] = chi_sq / 3

            fit_df.append(temp_df)

            # store curve for plotting
            temp3_df = dict(zip(x_line, ll3(x_line, *popt)))
            temp3_df[grouping_variable] = name
            fitted_df.append(temp3_df)


        # print the model parameters and chi2 to manually compare groups
        fit_df = pd.DataFrame(fit_df).set_index(grouping_variable)
        fit_df = fit_df.drop(columns=['slope', 'maximum', 'emt50'])
        print(fit_df)

        
        fitted_df = pd.DataFrame(fitted_df)
        fitted_df = pd.melt(fitted_df, id_vars=grouping_variable, 
                        value_vars=fitted_df.loc[:, fitted_df.columns != grouping_variable], 
                        var_name=time, value_name='value')

        
        # plot the observed data as points and the fitted models as curves
        sns.lmplot(data=self.bioassay, x=time, y='relative_stage', hue = grouping_variable, fit_reg=False, palette="colorblind")
        sns.lineplot(data=fitted_df, x=time, y='value', hue=grouping_variable, palette="colorblind")
        
        self.bioassay = self.bioassay.drop(columns='relative_stage')

# Next: - add statistics to compare groups and return p-values
#       - return sugestion for resistant/susceptible grouping
#       - return df with resistant/susceptible grouping
#       - add more model options
#       - option for choosing model based on best fit
#       - option for prediction if curve is not finished


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['phloemfinder']

package_data = \
{'': ['*']}

install_requires = \
['TPOT>=0.11.7', 'pandas>=1.5.1', 'scikit-learn>=0.24.1', 'seaborn>=0.12.1']

setup_kwargs = {
    'name': 'phloemfinder',
    'version': '0.3.1',
    'description': 'Find plant metabolites related to whitefly pest resistance',
    'long_description': '# phloemfinder\n\nFind plant metabolites related to whitefly pest resistance\n\n## Installation\n\n```bash\n$ pip install phloemfinder\n```\n\n## Usage\n\n- TODO\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`phloemfinder` was created by Lissy-Anne Denkers and Marc Galland. It is licensed under the terms of the Apache License 2.0 license.\n\n## Credits\n\n`phloemfinder` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n\n## Useful reading\n\n- [Autosklearn talks](https://github.com/automl/auto-sklearn-talks)\n- [Numpy docstring examples](https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_numpy.html#example-numpy)',
    'author': 'Marc Galland',
    'author_email': 'm.galland@uva.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/BleekerLab/phloemfinder',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)



#!/usr/bin/env python3 

import os
from warnings import WarningMessage
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


from sklearn.model_selection import train_test_split, cross_val_score
from sklearn.ensemble import RandomForestClassifier
from sklearn.pipeline import make_pipeline
from sklearn.inspection import permutation_importance

from autosklearn.classification import AutoSklearnClassifier
from autosklearn.metrics import balanced_accuracy, precision, recall, f1

from tpot import TPOTClassifier
from tpot.export_utils import set_param_recursive

from phloemfinder.utils import compute_metrics_classification 

class MetabolitePhenotypeFeatureSelection:
    '''
    A class to perform metabolite feature selection using phenotyping and metabolic data. 

    - Perform sanity checks on input dataframes (values above 0, etc.).
    - Get a baseline performance of a simple Machine Learning Random Forest ("baseline").
    - Perform automated Machine Learning model selection using autosklearn.
        Using metabolite data, train a model to predict phenotypes.
        Yields performance metrics (balanced accuracy, precision, recall) on the selected model.
    - Extracts performance metrics from the best ML model. 
    - Extracts the best metabolite features based on their feature importance and make plots per sample group. 

    Parameters
    ----------
    metabolome_csv: string
        A path to a .csv file with the cleaned up metabolome data (unreliable features filtered out etc.)
        Use the MetabolomeAnalysis class methods. 
        Shape of the dataframe is usually (n_samples, n_features) with n_features >> n_samples
    phenotype_csv: string
        A path to a .csv file with the phenotyping data. 
        Should be two columns at least with: 
          - column 1 containing the sample identifiers
          - column 2 containing the phenotypic class e.g. 'resistant' or 'sensitive'
    metabolome_feature_id_col: string, default='feature_id'
        The name of the column that contains the feature identifiers.
        Feature identifiers should be unique (=not duplicated).
    phenotype_sample_id: string, default='sample_id'
        The name of the column that contains the sample identifiers.
        Sample identifiers should be unique (=not duplicated).


    Attributes
    ----------
    metabolome_validated: bool
      Is the metabolome file valid for Machine Learning? (default is False)   

    phenotype_validated: bool
      Is the phenotype file valid for Machine Learning? (default is False)

    baseline_performance: float 
      The baseline performance computed with get_baseline_performance() i.e. using a simple Random Forest model. 
      Search for the best ML model using search_best_model() should perform better than this baseline performance. 

    best_ensemble_models_searched: bool
      Is the search for best ensemble model using auto-sklearn already performed? (default is False)

    metabolome: pandas.core.frame.DataFrame
      The validated metabolome dataframe of shape (n_features, n_samples).
    
    phenotype: pandas.core.frame.DataFrame
      A validated phenotype dataframe of shape (n_samples, 1)
      Sample names in the index and one column named 'phenotype' with the sample classes.
    
    baseline_performance: str
      Average balanced accuracy score (-/+ standard deviation) of the basic Random Forest model. 
    
    best_model: sklearn.pipeline.Pipeline
      A scikit-learn pipeline that contains one or more steps.
      It is the best performing pipeline found by TPOT automated ML search.

    feature_importances: pandas.core.frame.DataFrame
       A Pandas dataframe that contains feature importance information using scikit-learn permutation_importance()
        Mean of feature importance over n_repeats.
        Standard deviation over n_repeats.
        Raw permutation importance scores.
    
    
    Methods
    --------
    validate_input_metabolome_df()
      Validates the dataframe read from the 'metabolome_csv' input file.
    
    validate_input_phenotype_df()
      Validates the phenotype dataframe read from the 'phenotype_csv' input file.
    
    get_baseline_performance()
      Fits a basic Random Forest model to get default performance metrics. 
    
    search_best_model_with_tpot_and_get_feature_importances()
      Search for the best ML pipeline using TPOT genetic programming method.
      Computes and output performance metrics from the best pipeline.
      Extracts feature importances using scikit-learn permutation_importance() method. 

    

    Notes
    --------

    Example of an input metabolome .csv file

        | feature_id  | genotypeA_rep1 | genotypeA_rep2 | genotypeA_rep3 | genotypeA_rep4 |
        |-------------|----------------|----------------|----------------|----------------|
        | metabolite1 |   1246         | 1245           | 12345          | 12458          |
        | metabolite2 |   0            | 0              | 0              | 0              |
        | metabolite3 |   10           | 0              | 0              | 154            |

    Example of an input phenotype .csv file

        | sample_id      | phenotype | 
        |----------------|-----------|
        | genotypeA_rep1 | sensitive | 
        | genotypeA_rep2 | sensitive |   
        | genotypeA_rep3 | sensitive |
        | genotypeA_rep4 | sensitive | 
        | genotypeB_rep1 | resistant |   
        | genotypeB_rep2 | resistant |
    
    '''
    # Class attribute shared among all instances of the class
    # By default the metabolome and phenotype data imported from .csv files will have to be validated
    # By default all filters have not been executed (blank filtering, etc.)
    # Baseline performance of a simple ML model as well as search of best model are also None/False by default. 
    metabolome_validated=False
    phenotype_validated=False
    baseline_performance=None
    best_ensemble_models_searched=False

    # Class constructor method
    def __init__(
        self, 
        metabolome_csv, 
        phenotype_csv,
        metabolome_feature_id_col='feature_id', 
        phenotype_sample_id='sample_id'):
        
        # Import metabolome dataframe and verify presence of feature id column
        self.metabolome = pd.read_csv(metabolome_csv)
        if metabolome_feature_id_col not in self.metabolome.columns:
            raise ValueError("The specified column with feature identifiers '{0}' is not present in your '{1}' file.".format(metabolome_feature_id_col,os.path.basename(metabolome_csv)))
        else:
            self.metabolome.set_index(metabolome_feature_id_col, inplace=True)

        # Import phenotype dataframe and verify presence of sample id column
        self.phenotype = pd.read_csv(phenotype_csv)
        if phenotype_sample_id not in self.phenotype.columns:
            raise ValueError("The specified column with sample identifiers '{0}' is not present in your '{1}' file.".format(phenotype_sample_id, os.path.basename(phenotype_csv)))
        else:
            try: 
                self.phenotype.set_index(phenotype_sample_id, inplace=True)
            except:
                raise IndexError("Values for sample identifiers have to be unique. Check your ", phenotype_sample_id, " column.")

    ################
    ## Verify inputs
    ################
    def validate_input_metabolome_df(self):
        '''
        Validates the dataframe containing the feature identifiers, metabolite values and sample names.
        Will place the 'feature_id_col' column as the index of the validated dataframe. 
        The validated metabolome dataframe is stored as the 'validated_metabolome' attribute 
        
        
        Returns
        --------
        self: object
          Object with metabolome_validated set to True

        Example of a validated output metabolome dataframe

                      | genotypeA_rep1 | genotypeA_rep2 | genotypeA_rep3 | genotypeA_rep4 |
                      |----------------|----------------|----------------|----------------|
          feature_id
        | metabolite1 |   1246         | 1245           | 12345          | 12458          |
        | metabolite2 |   0            | 0              | 0              | 0              |
        | metabolite3 |   10           | 0              | 0              | 154            |
        '''
        
        if np.any(self.metabolome.values < 0):
            raise ValueError("Sorry, metabolite values have to be zero or positive integers (>=0)")
        else:
            self.metabolome_validated = True
            print("Metabolome data validated.")
    
    def validate_input_phenotype_df(self, phenotype_class_col="phenotype"):
        '''
        Validates the dataframe containing the phenotype classes and the sample identifiers

        Params
        ------
        phenotype_class_col: string, default="phenotype"
            The name of the column to be used 

        Returns
        --------
        self: object
          Object with phenotype_validated set to True

        Notes
        --------
        Example of an input phenotype dataframe
        

        | sample_id      | phenotype | 
        |----------------|-----------|
        | genotypeA_rep1 | sensitive | 
        | genotypeA_rep2 | sensitive |   
        | genotypeA_rep3 | sensitive |
        | genotypeA_rep4 | sensitive | 
        | genotypeB_rep1 | resistant |   
        | genotypeB_rep2 | resistant |

        Example of a validated output phenotype dataframe. 

                         | phenotype | 
                         |-----------|
          sample_id      
        | genotypeA_rep1 | sensitive | 
        | genotypeA_rep2 | sensitive |   
        | genotypeA_rep3 | sensitive |
        | genotypeA_rep4 | sensitive | 
        | genotypeB_rep1 | resistant |   
        | genotypeB_rep2 | resistant |

        Example
        -------
        >> fs = MetabolitePhenotypeFeatureSelection(
        >>        metabolome_csv="clean_metabolome.csv", 
        >>        phenotype_csv="phenotypes_test_data.csv", 
        >>        phenotype_sample_id='sample')
        >> fs.validate_input_phenotype_df()

        '''
        n_distinct_classes = self.phenotype[phenotype_class_col].nunique()
        try:
            n_distinct_classes == 2
            self.phenotype_validated = True    
            print("Phenotype data validated.")
        except:
            raise ValueError("The number of distinct phenotypic classes in the {0} column should be exactly 2.".format(phenotype_class_col))
    
    #################
    ## Baseline model
    #################
    def get_baseline_performance(
      self, 
      class_of_interest,
      kfold=5, 
      train_size=0.8,
      random_state=123,
      scoring_metric='balanced_accuracy'):
        '''
        Takes the phenotype and metabolome dataset and compute a simple Random Forest analysis with default hyperparameters. 
        This will give a base performance for a Machine Learning model that has then to be optimised using autosklearn

        k-fold cross-validation is performed to mitigate split effects on small datasets. 

        Parameters
        ----------
        class_of_interest: str
          The name of the class of interest also called "positive class".
          This class will be used to calculate recall_score and precision_score. 
          Recall score = TP / (TP + FN) with TP: true positives and FN: false negatives.
          Precision score = TP / (TP + FP) with TP: true positives and FP: false positives. 

        kfold: int, optional
          Cross-validation strategy. Default is to use a 5-fold cross-validation. 
        
        train_size: float or int, optional
          If float, should be between 0.5 and 1.0 and represent the proportion of the dataset to include in the train split.
          If int, represents the absolute number of train samples. If None, the value is automatically set to the complement of the test size.
          Default is 0.8 (80% of the data used for training).

        random_state: int, optional
          Controls both the randomness of the train/test split  samples used when building trees (if bootstrap=True) and the sampling of the features to consider when looking for the best split at each node (if max_features < n_features). See Glossary for details.
          You can change this value several times to see how it affects the best ensemble model performance.
          Default is 123.


        scoring_metric: str, optional
          A valid scoring value (default="balanced_accuracy")
          To get a complete list, type:
          >> from sklearn.metrics import SCORERS 
          >> sorted(SCORERS.keys()) 
          balanced accuracy is the average of recall obtained on each class. 

        Returns
        -------
        self: object
          Object with baseline_performance attribute.
        
        Example
        -------
        >>> fs = MetabolitePhenotypeFeatureSelection(
                   metabolome_csv="../tests/clean_metabolome.csv", 
                   phenotype_csv="../tests/phenotypes_test_data.csv", 
                   phenotype_sample_id='sample')
            fs.get_baseline_performance()

        '''
        try:
            self.metabolome_validated == True
        except:
            raise ValueError("Please validate metabolome data first using the validate_input_metabolome_df() method.")

        try:
            self.phenotype_validated == True
        except:
            raise ValueError("Please validate phenotype data first using the validate_input_phenotype_df() method.")

        X = self.metabolome.transpose()
        y = self.phenotype.values.ravel() # ravel makes the array contiguous
        X_train, X_test, y_train, y_test = train_test_split(
          X, y, 
          train_size=train_size, 
          random_state=random_state, 
          stratify=y)

        # Train model and assess performance
        clf = RandomForestClassifier(n_estimators=1000, random_state=random_state)
        scores = cross_val_score(clf, X_train, y_train, scoring=scoring_metric, cv=kfold)
        average_scores = np.average(scores).round(3) * 100
        stddev_scores = np.std(scores).round(3) * 100
        print("====== Training a basic Random Forest model =======")
        baseline_performance = "Average {0} score on training data is: {1:.3f} % -/+ {2:.2f}".format(scoring_metric, average_scores, stddev_scores)
        print(baseline_performance)
        print("\n")
        print("====== Performance on test data of the basic Random Forest model =======")
        clf.fit(X_train, y_train)
        predictions = clf.predict(X_test)
        model_balanced_accuracy_score = balanced_accuracy(y_true=y_test, y_pred=predictions).round(3) * 100 
        print("Average {0} score on test data is: {1:.3f} %".format(scoring_metric, model_balanced_accuracy_score))
        self.baseline_performance = baseline_performance

    # TODO: make decorator function to check arguments     
    # See -> https://www.pythonforthelab.com/blog/how-to-use-decorators-to-validate-input/
    def search_best_model_with_tpot_and_get_feature_importances(
        self,
        class_of_interest,
        scoring_metric='balanced_accuracy',
        kfolds=3,
        train_size=0.8,
        max_time_mins=5,
        max_eval_time_mins=1,
        random_state=123,
        tpot_config_dict=None,
        n_permutations=10):
      '''
      Search for the best ML model with TPOT genetic programming methodology and extracts feature importances. 

      A resampling strategy called "cross-validation" will be performed on a subset of the data (training data) to increase 
      the model generalisation performance. Finally, the model performance is tested on the unseen test data subset.  
      
      By default, TPOT will make use of a set of preprocessors (e.g. Normalizer, PCA) and algorithms (e.g. RandomForestClassifier)
      defined in the default config (classifier.py).
      See: https://github.com/EpistasisLab/tpot/blob/master/tpot/config/classifier.py

      Parameters
      ----------
      class_of_interest: str
          The name of the class of interest also called "positive class".
          This class will be used to calculate recall_score and precision_score. 
          Recall score = TP / (TP + FN) with TP: true positives and FN: false negatives.
          Precision score = TP / (TP + FP) with TP: true positives and FP: false positives. 

      scoring_metric: str, optional
        Function used to evaluate the quality of a given pipeline for the classification problem. 
        Default is 'balanced accuracy'. 
        The following built-in scoring functions can be used:
          'accuracy', 'adjusted_rand_score', 'average_precision', 'balanced_accuracy', 
          'f1', 'f1_macro', 'f1_micro', 'f1_samples', 'f1_weighted', 'neg_log_loss', 
          'precision' etc. (suffixes apply as with ‘f1’), 'recall' etc. (suffixes apply as with ‘f1’), 
          ‘jaccard’ etc. (suffixes apply as with ‘f1’), 'roc_auc', ‘roc_auc_ovr’, ‘roc_auc_ovo’, ‘roc_auc_ovr_weighted’, ‘roc_auc_ovo_weighted’ 

      kfolds: int, optional
        kfolds: int, optional
        Number of folds for the stratified K-Folds cross-validation strategy. Default is 3-fold cross-validation. 
        Has to be comprised between 3 and 10 i.e. 3 <= kfolds =< 10
        See https://scikit-learn.org/stable/modules/cross_validation.html
      
      train_size: float or int, optional
        If float, should be between 0.5 and 1.0 and represent the proportion of the dataset to include in the train split.
        If int, represents the absolute number of train samples. If None, the value is automatically set to the complement of the test size.
        Default is 0.8 (80% of the data used for training).

      max_time_mins: int, optional
        How many minutes TPOT has to optimize the pipeline (in total). Default is 5 minutes.
        This setting will allow TPOT to run until max_time_mins minutes elapsed and then stop.
        Try short time intervals (5, 10, 15min) and then see if the model score on the test data improves. 
      
      max_eval_time_mins: float, optional 
        How many minutes TPOT has to evaluate a single pipeline. Default is 1min. 
        This time has to be smaller than the 'max_time_mins' setting.

      random_state: int, optional
          Controls both the randomness of the train/test split  samples used when building trees (if bootstrap=True) and the sampling of the features to consider when looking for the best split at each node (if max_features < n_features). See Glossary for details.
          You can change this value several times to see how it affects the best ensemble model performance.
          Default is 123.

      n_permutations: int, optional
        Number of permutations used to compute feature importances from the best model using scikit-learn permutation_importance() method.
        Default is 10 permutations.
      
      tpot_config_dict: Python dictionary, str or None, optional.
        A configuration dictionary for customizing the operators and parameters that TPOT searches. 
        Default is None meaning that TPOT will buse the default configuration. See https://epistasislab.github.io/tpot/using/#customizing-tpots-operators-and-parameters 
        A customized Python dictionary can be made and passed to TPOTClassifier. See https://epistasislab.github.io/tpot/using/#customizing-tpots-operators-and-parameters
       

       Returns
        ------
        self: object
          The object with best model searched and feature importances computed. 

       See also
       --------

       Notes
       -----
       Permutation importances are calculated on the training set
       Permutation importances can be computed either on the training set or on a held-out testing or validation set.
       Using a held-out set makes it possible to highlight which features contribute the most to the generalization power of the inspected model. 
       Features that are important on the training set but not on the held-out set might cause the model to overfit.
       https://scikit-learn.org/stable/modules/permutation_importance.html#permutation-importance 
          
      '''
      X = self.metabolome.transpose().to_numpy(dtype='float64')
      y = self.phenotype.values.ravel()
        
      # Verify input arguments
      try:
        3 <= kfolds <= 10
      except:
        raise ValueError('kfolds argument has to be comprised between 3 and 10')
        
      try:
        0.5 < train_size < 0.9
      except:
        raise ValueError('train_size has to be comprised between 0.5 and 0.9')

      try:
        class_of_interest in set(y)
      except:
        raise ValueError('The class_of_interest value "{0}" has to be in the phenotype labels {1}'.format(class_of_interest, set(y)))
      
      ### Automated search for best model/pipeline
      X_train, X_test, y_train, y_test = train_test_split(
          X, y, 
          train_size=train_size, 
          random_state=random_state, 
          stratify=y)
      tpot = TPOTClassifier(
        generations=None, # If None, the parameter max_time_mins must be defined as the runtime limit. 
        max_time_mins=max_time_mins,
        max_eval_time_mins=max_eval_time_mins,
        cv=kfolds,        # Number of folds in an unshuffled StratifiedKFold.
        config_dict=tpot_config_dict,
        random_state=random_state, 
        verbosity=0)
      tpot.fit(X_train, y_train)
      best_pipeline = make_pipeline(tpot.fitted_pipeline_)
      
      ### Fit best model/pipeline
      if len(best_pipeline) == 1:
        # one step pipeline thus one model  
        one_step_pipeline = best_pipeline["pipeline"][-1]
        one_step_pipeline.fit(X_train, y_train)
        print(one_step_pipeline.score(X_test, y_test))
        predictions = one_step_pipeline.predict(X_test)
        print("============ Performance of ML model on train data =============")
        print("Train {0} score {1:.3f}".format(scoring_metric, one_step_pipeline.score(X_train, y_train).round(3)*100))
        print("\n")
        print("============ Performance of ML model on test data =============")
        print(compute_metrics_classification(y_predictions=predictions, y_trues=y_test, positive_class=class_of_interest))
        self.best_model = one_step_pipeline
      else:
        # multiple steps pipeline
        n_steps_pipeline = make_pipeline(tpot.fitted_pipeline_)
        set_param_recursive(best_pipeline, 'random_state', random_state)
        n_steps_pipeline.fit(X_train, y_train)
        predictions = n_steps_pipeline.predict(X_test)
        training_score = n_steps_pipeline.score(X_train, y_train).round(3) * 100
        print("============ Performance of ML model on train data =============")
        print("Train {0} score {1:.3f}".format(scoring_metric, training_score))
        print("\n")
        print("============ Performance of ML model on test data =============")
        print(compute_metrics_classification(y_predictions=predictions, y_trues=y_test, positive_class=class_of_interest))
        self.best_model = n_steps_pipeline

      ### Compute feature importances
      # Has to be done on the same train/test split. 
      print("\n")
      print("======== Computing feature importances on the training set =======")
      feature_importances_training_set = permutation_importance(
        self.best_model, 
        X=X_train, 
        y=y_train, 
        scoring=scoring_metric, 
        n_repeats=n_permutations, 
        random_state=random_state)
      mean_imp = pd.DataFrame(feature_importances_training_set.importances_mean, columns=["mean_var_imp"])
      std_imp = pd.DataFrame(feature_importances_training_set.importances_std, columns=["std_var_imp"])
      raw_imp = pd.DataFrame(feature_importances_training_set.importances, columns=["perm" + str(i) for i in range(n_permutations)])

      feature_importances = pd.concat([mean_imp, std_imp], axis=1).set_index(self.metabolome.index.values).sort_values('mean_var_imp', ascending=False)

      self.feature_importances = feature_importances
      self.feature_importances_permutations = raw_imp